#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

try:
    from lxml import etree
    from lxml.builder import E
except ImportError:
    pass

csa_tag = 'CSA'
csa_namespace = 'http://software.incf.org/software/csa/1.0'
CSA = '{%s}' % csa_namespace

CUSTOM = -4
OPERATOR = -3
BINDOPERATOR = -2
SINGLETON = -1

def to_xml (obj):
    if isinstance (obj, str):
        return E (obj)
    elif isinstance (obj, (int, float)):
        return E ('cn', str (obj))
    elif isinstance (obj, CSAObject):
        return obj._to_xml ()
    else:
        raise RuntimeError, "don't know how to turn %s into xml" % obj

# precedence levels:
#
# 0 + -
# 1 *
# 2 ~

class CSAObject (object):
    tag_map = {}

    def __init__ (self, name, precedence = 3):
        self.name = name
        self.precedence = precedence

    def __repr__ (self):
        return 'CSA(%s)' % self.repr ()

    def repr (self):
        if hasattr (self, 'name'):
            return self.name
        else:
            return self.__class__.__name__

    def _repr_as_op2 (self, parentPrecedence):
        if self.precedence <= parentPrecedence:
            return '(%s)' % self.repr ()
        else:
            return self.repr ()

    def _repr_applyop (self, op_repr, obj):
        return '%s*%s' % (op_repr, obj._repr_as_op2 (1))

    def to_xml (self):
        return E (csa_tag, self._to_xml (), xmlns=csa_namespace)

    def _to_xml (self):
        return E (self.name)

    @classmethod
    def apply (cls, operator, *operands):
        return E ('apply', to_xml (operator), *map (to_xml, operands))

    @classmethod
    def formalFromXML (cls, element):
        assert element.tag == CSA + 'bvar'
        nodes = element.getchildren ()
        assert nodes[0].tag == CSA + 'ci'
        return nodes[0].text

    @classmethod
    def from_xml (cls, element, env = {}):
        if element.tag == CSA + 'cn':
            return eval (element.text)
        elif element.tag == CSA + 'ci':
            #*fixme* Implement env as lists of dictionaries
            return env[element.text]
        elif element.tag == CSA + 'apply':
            nodes = element.getchildren ()
            operator = nodes[0].tag
            operands = [ cls.from_xml (e, env) for e in nodes[1:] ]
            if operator == CSA + 'plus':
                return operands[0].__add__ (operands[1])
            elif operator == CSA + 'minus':
                return operands[0].__sub__ (operands[1])
            elif operator == CSA + 'times':
                return operands[0].__mul__ (operands[1])
            elif operator == CSA + 'complement':
                return operands[0].__invert__ ()
            else:
                # Function or operator application
                entry = CSAObject.tag_map[operator]
                obj = entry[0]
                if entry[1] == OPERATOR:
                    return obj * operands[1]
                else:
                    return obj (*operands)
        elif element.tag == CSA + 'bind':
            nodes = element.getchildren ()
            tag = nodes[0].tag
            entry = CSAObject.tag_map[tag]
            if entry[1] != BINDOPERATOR:
                raise RuntimeError, "unknown binding operator tag %s" % tag
            bindingOperator = entry[0]
            bvars = [ CSAObject.formalFromXML (e) for e in nodes[1:-1] ]
            return bindingOperator (bvars, nodes[-1])
        elif element.tag in CSAObject.tag_map:
            entry = CSAObject.tag_map[element.tag]
            obj = entry[0]
            if entry[1] == SINGLETON:
                return obj
            elif entry[1] == CUSTOM:
                return obj.from_xml (element, env)
            else:
                return obj ()
        else:
            raise RuntimeError, "don't know how parse tag %s" % element.tag

    def xml (e):
        print etree.tostring (e)

    def write(self, file):
        doc = self.to_xml ()
        etree.ElementTree(doc).write (file, encoding="UTF-8",
                                      pretty_print=True, xml_declaration=True)

class BinaryCSAObject (CSAObject):
    operator_table = {'+': 'plus', '-': 'minus', '*': 'times'}
    
    def __init__ (self, name, op1, op2, precedence = 0):
        CSAObject.__init__ (self, name, precedence)
        self.op1 = op1
        self.op2 = op2

    def repr (self):
        if isinstance (self.op1, CSAObject):
            op1 = self.op1.repr ()
            if self.op1.precedence < self.precedence:
                op1 = "(%s)" % op1
        else:
            op1 = self.op1

        if isinstance (self.op2, CSAObject):
            op2 = self.op2._repr_as_op2 (self.precedence)
        else:
            op2 = self.op2
            
        return "%s%s%s" % (op1, self.name, op2)

    def _to_xml (self):
        if isinstance (self.op1, CSAObject):
            op1 = self.op1._to_xml ()
        else:
            op1 = self.op1

        if isinstance (self.op2, CSAObject):
            op2 = self.op2._to_xml ()
        else:
            op2 = self.op2

        if self.name in BinaryCSAObject.operator_table:
            op = BinaryCSAObject.operator_table[self.name]
        else:
            op = self.name
        return E ('apply', E (op), op1, op2)


class OpExprValue (BinaryCSAObject):
    def __init__ (self, operator, operand):
        BinaryCSAObject.__init__ (self, '*', operator, operand, 1)


class Operator (CSAObject):
    def __init__ (self, name='ioperator'):
        CSAObject.__init__ (self, name)


def from_xml (root):
    assert root.nsmap[None] == csa_namespace
    return CSAObject.from_xml (root.getchildren ()[0])


def parse (filename):
    doc = etree.parse (filename)
    return from_xml (doc.getroot())


def registerTag (tag, obj, mode):
    CSAObject.tag_map[CSA + tag] = (obj, mode)
 



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from csaobject import *

import inspect

try:
    from lxml import etree
    from lxml.builder import E
except ImportError:
    pass

class Closure (CSAObject):
    tag = 'closure'
    name = tag
    
    def __init__ (self, formals, e):
        self.formals = formals
        self.etree = e

    @staticmethod
    def formalToXML (formal):
        return E ('bvar', E ('ci', formal))
    
    def _to_xml (self):
        formals = map (Closure.formalToXML, self.formals)
        return E ('bind', E ('closure'), *formals + [ self.etree ])

    def __call__ (self, *args):
        assert len (args) == len (self.formals), "arguments %s don't match formals %s" % (args, self.formals)
        bindings = {}
        for (formal, arg) in map (None, self.formals, args):
            bindings[formal] = arg
        return CSAObject.from_xml (self.etree, bindings)

registerTag (Closure.tag, Closure, BINDOPERATOR)



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from csaobject import *

class ValueSet (CSAObject):
    def __init__ (self):
        CSAObject.__init__ (self, "valueset")
        
    def __neg__ (self):
        return GenericValueSet (lambda i, j: - self (i, j))
    
    def __add__ (self, other):
        if not callable (other):
            return maybeAffine (other, 1.0, self)
        elif isinstance (other, (QuotedValueSet, AffineValueSet)):
            return other.__add__ (self)
        elif isinstance (other, GenericValueSet):
            return GenericValueSet (lambda i, j: self (i, j) + other.function (i, j))
        else:
            return GenericValueSet (lambda i, j: self (i, j) + other (i, j))

    def __radd__ (self, other):
        return self.__add__ (other)

    def __sub__ (self, other):
        return self.__add__ (- other)

    def __rsub__ (self, other):
        return self.__neg__ ().__add__ (other)

    def __mul__ (self, other):
        if not callable (other):
            return maybeAffine (0.0, other, self)
        elif isinstance (other, (QuotedValueSet, AffineValueSet)):
            return other.__mul__ (self)
        elif isinstance (other, GenericValueSet):
            return GenericValueSet (lambda i, j: self (i, j) * other.function (i, j))
        else:
            return GenericValueSet (lambda i, j: self (i, j) * other (i, j))

    def __rmul__ (self, other):
        return self.__mul__ (other)


class QuotedValueSet (ValueSet):
    def __init__ (self, expression):
        ValueSet.__init__ (self)
        self.expression = expression

    def __call__ (self, i, j):
        return self.expression

    def __neg__ (self):
        return QuotedValueSet (- self.expression)
    
    def __add__ (self, other):
        if not callable (other):
            return QuotedValueSet (self.expression + other)
        elif isinstance (other, QuotedValueSet):
            return QuotedValueSet (self.expression + other.expression)
        elif isinstance (other, AffineValueSet):
            return other.__add__ (self)
        else:
            return maybeAffine (self.expression, 1.0, other)

    def __mul__ (self, other):
        if not callable (other):
            return QuotedValueSet (self.expression * other)
        elif isinstance (other, QuotedValueSet):
            return QuotedValueSet (self.expression * other.expression)
        elif isinstance (other, AffineValueSet):
            return other.__mul__ (self)
        else:
            return maybeAffine (0.0, self.expression, other)        


class GenericValueSet (ValueSet):
    def __init__ (self, function):
        ValueSet.__init__ (self)
        self.function = function

    def __call__ (self, i, j):
        return self.function (i, j)

    def __neg__ (self):
        return GenericValueSet (lambda i, j: - self.function (i, j))

    def __add__ (self, other):
        if not callable (other):
            return maybeAffine (other, 1.0, self.function)
        elif isinstance (other, (QuotedValueSet, AffineValueSet)):
            return other.__add__ (self)
        elif isinstance (other, GenericValueSet):
            return GenericValueSet (lambda i, j: self.function (i, j) + other.function (i, j))
        else:
            return GenericValueSet (lambda i, j: self.function (i, j) + other (i, j))

    def __mul__ (self, other):
        if not callable (other):
            return maybeAffine (0.0, other, self.function)
        elif isinstance (other, (QuotedValueSet, AffineValueSet)):
            return other.__mul__ (self)
        elif isinstance (other, GenericValueSet):
            return GenericValueSet (lambda i, j: self.function (i, j) * other.function (i, j))
        else:
            return GenericValueSet (lambda i, j: self.function (i, j) * other (i, j))


class AffineValueSet (ValueSet):
    def __init__ (self, constant, coefficient, function):
        ValueSet.__init__ (self)
        self.const = constant
        self.coeff = coefficient
        self.func = function

    def __call__ (self, i, j):
        return self.const + self.coeff * self.func (i, j)

    def __neg__ (self):
        return maybeAffine (- self.const, - self.coeff, self.func)
    
    def __add__ (self, other):
        if not callable (other):
            return maybeAffine (self.const + other, self.coeff, self.func)
        elif isinstance (other, QuotedValueSet):
            return maybeAffine (self.const + other.expression,
                                self.coeff, self.func)
        elif isinstance (other, AffineValueSet):
            f = lambda i, j: \
                    self.const * self.func (i, j) \
                    + other.const * other.func (i, j)
            return maybeAffine (self.const + other.const,
                                1.0,
                                f)

    def __mul__ (self, other):
        if not callable (other):
            return maybeAffine (self.const * other,
                                self.coeff * other,
                                self.func)
        elif isinstance (other, QuotedValueSet):
            return maybeAffine (self.const * other.expression,
                                self.coeff * other.expression,
                                self.func)
        elif isinstance (other, AffineValueSet):
            f = lambda i, j: \
                    other.const * self.coeff * self.func (i, j) \
                    + self.const * other.coeff * other.func (i, j) \
                    + self.coeff * other.coeff \
                      * self.func (i, j) * other.func (i, j)
            return maybeAffine (self.const * other.const,
                                1.0,
                                f)

def maybeAffine (const, coeff, func):
    if coeff == 0.0:
        return QuotedValueSet (const)
    elif const == 0.0 and coeff == 1.0:
        return GenericValueSet (func)
    else:
        return AffineValueSet (const, coeff, func)



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import random
import numpy
import copy

import connset as cs
import intervalset as iset

from csaobject import *

class FullMask (cs.IntervalSetMask):
    tag = 'full'
    
    def __init__ (self):
        cs.IntervalSetMask.__init__ (self, iset.N, iset.N)
        self.name = FullMask.tag
        CSAObject.tag_map[CSA + FullMask.tag] = (self, SINGLETON)

    def __call__ (self, N0, N1 = None):
        if N1 == None:
            N1 = N0
        return cs.FiniteISetMask (iset.IntervalSet ((0, N0 - 1)), \
                                  iset.IntervalSet ((0, N1 - 1)))


class OneToOne (cs.Mask):
    tag = 'oneToOne'
    
    def __init__ (self):
        cs.Mask.__init__ (self)
        self.name = OneToOne.tag
        CSAObject.tag_map[CSA + OneToOne.tag] = (self, SINGLETON)
    
    def iterator (self, low0, high0, low1, high1, state):
        for i in xrange (max (low0, low1), min (high0, high1)):
            yield (i, i)


class ConstantRandomMask (cs.Mask):
    tag = 'randomMask'
    
    def __init__ (self, p):
        cs.Mask.__init__ (self)
        self.p = p
        self.state = random.getstate ()
        self.name = ConstantRandomMask.tag

    def startIteration (self, state):
        random.setstate (self.state)
        return self

    def iterator (self, low0, high0, low1, high1, state):
        for j in xrange (low1, high1):
            for i in xrange (low0, high0):
                if random.random () < self.p:
                    yield (i, j)

    def repr (self):
        return 'random(%s)' % self.p

    def _to_xml (self):
        return CSAObject.apply (ConstantRandomMask.tag, self.p)

registerTag (ConstantRandomMask.tag, ConstantRandomMask, 1)


class SampleNRandomOperator (cs.Operator):
    tag = 'random_N'
    
    def __init__ (self, N):
        self.N = N

    def __mul__ (self, other):
        assert isinstance (other, cs.Finite) \
               and isinstance (other, cs.Mask), \
               'expected finite mask'
        return SampleNRandomMask (self.N, other)

    def repr (self):
        return 'random(N = %s)' % self.N

    def _to_xml (self):
        return CSAObject.apply (SampleNRandomOperator.tag, self.N)

registerTag (SampleNRandomOperator.tag, SampleNRandomOperator, 1)


class SampleNRandomMask (cs.Finite,cs.Mask):
    # The algorithm based on first sampling the number of connections
    # per partition has been arrived at through discussions with Hans
    # Ekkehard Plesser.
    #
    def __init__ (self, N, mask):
        cs.Mask.__init__ (self)
        self.N = N
        assert isinstance (mask, cs.FiniteISetMask), \
               'SampleNRandomMask currently only operates on FiniteISetMask:s'
        self.mask = mask
        self.randomState = random.getstate ()
        self.npRandomState = numpy.random.get_state ()

    def bounds (self):
        return self.mask.bounds ()

    def startIteration (self, state):
        obj = copy.copy (self)  # local state: N, N0, perTarget, sources
        random.setstate (self.randomState)
        obj.isPartitioned = False
        if 'partitions' in state:
            obj.isPartitioned = True
            partitions = map (self.mask.intersection, state['partitions'])
            sizes = map (len, partitions)
            total = sum (sizes)
            
            # The following yields the same result on all processes.
            # We should add a seed function to the CSA.
            if 'seed' in state:
                seed = state['seed']
            else:
                seed = 'SampleNRandomMask'
            numpy.random.seed (hash (seed))
            
            N = numpy.random.multinomial (self.N, numpy.array (sizes) \
                                           / float (total))
            obj.N = N[state['selected']]
            obj.mask = partitions[state['selected']]
            assert isinstance (obj.mask, cs.FiniteISetMask), \
                   'SampleNRandomMask iterator only handles finite IntervalSetMask partitions'
        obj.mask = obj.mask.startIteration (state)
        obj.N0 = len (obj.mask.set0)
        obj.lastBound0 = False
        N1 = len (obj.mask.set1)
        numpy.random.set_state (self.npRandomState)
        obj.perTarget = numpy.random.multinomial (obj.N, [1.0 / N1] * N1)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        m = self.mask.set1.count (0, low1)
        
        if self.isPartitioned and m > 0:
            # "replacement" for a proper random.jumpahead (n)
            # This is required so that different partitions of this
            # mask aren't produced using the same stream of random
            # numbers.
            random.seed (random.getrandbits (32) + m)
            
        if self.lastBound0 != (low0, high0):
            self.lastBound0 = (low0, high0)
            self.sources = []
            for i in self.mask.set0.boundedIterator (low0, high0):
                self.sources.append (i)

        nSources = len (self.sources)
        for j in self.mask.set1.boundedIterator (low1, high1):
            s = []
            for k in xrange (0, self.perTarget[m]):
                i = random.randint (0, self.N0 - 1)
                if i < nSources:
                    s.append (self.sources[i])
            s.sort ()
            for i in s:
                yield (i, j)
            m += 1

    def repr (self):
        return self._repr_applyop ('random(N=%s)' % self.N, self.mask)

    def _to_xml (self):
        return E ('apply',
                  E ('times'),
                  CSAObject.apply (SampleNRandomOperator.tag, self.N),
                  self.mask._to_xml ())


class FanInRandomOperator (cs.Operator):
    tag = 'random_fanIn'
    
    def __init__ (self, fanIn):
        self.fanIn = fanIn

    def __mul__ (self, other):
        assert isinstance (other, cs.Finite) \
               and isinstance (other, cs.Mask), \
               'expected finite mask'
        return FanInRandomMask (self.fanIn, other)

    def repr (self):
        return 'random(fanIn=%s)' % self.fanIn

    def _to_xml (self):
        return CSAObject.apply (FanInRandomOperator.tag, self.fanIn)

registerTag (FanInRandomOperator.tag, FanInRandomOperator, 1)


# This code is copied and modified from SampleNRandomMask
# *fixme* refactor code and eliminate code duplication
class FanInRandomMask (cs.Finite, cs.Mask):
    # The algorithm based on first sampling the number of connections
    # per partition has been arrived at through discussions with Hans
    # Ekkehard Plesser.
    #
    def __init__ (self, fanIn, mask):
        cs.Mask.__init__ (self)
        self.fanIn = fanIn
        assert isinstance (mask, cs.FiniteISetMask), \
               'FanInRandomMask currently only operates on FiniteISetMask:s'
        self.mask = mask
        self.randomState = random.getstate ()

    def bounds (self):
        return self.mask.bounds ()

    def startIteration (self, state):
        obj = copy.copy (self)  # local state: N, N0, perTarget, sources
        random.setstate (self.randomState)
        obj.isPartitioned = False
        if 'partitions' in state:
            obj.isPartitioned = True
            partitions = map (self.mask.intersection, state['partitions'])
            
            # The following yields the same result on all processes.
            # We should add a seed function to the CSA.
            if 'seed' in state:
                seed = state['seed']
            else:
                seed = 'FanInRandomMask'
            # Numpy.random.seed requires unsigned integer
            numpy.random.seed (abs (hash (seed)))

            selected = state['selected']
            obj.mask = partitions[selected]
            assert isinstance (obj.mask, cs.FiniteISetMask), \
                   'FanInRandomMask iterator only handles finite IntervalSetMask partitions'
        obj.mask = obj.mask.startIteration (state)
        obj.N0 = len (obj.mask.set0)
        obj.lastBound0 = False
        if obj.isPartitioned:
            obj.perTarget = []
            for j in obj.mask.set1:
                size = 0
                sourceDist = numpy.zeros (len (partitions))
                for k in xrange (len (partitions)):
                    if j in partitions[k].set1:
                        sourceDist[k] = len (partitions[k].set0)
                sourceDist /= sum (sourceDist)
                dist = numpy.random.multinomial (self.fanIn, sourceDist)
                obj.perTarget.append (dist[selected])
        else:
            obj.perTarget = [self.fanIn] * len (obj.mask.set1)
        return obj

    def iterator (self, low0, high0, low1, high1, state):
        m = self.mask.set1.count (0, low1)
        
        if self.isPartitioned and m > 0:
            # "replacement" for a proper random.jumpahead (n)
            # This is required so that different partitions of this
            # mask aren't produced using the same stream of random
            # numbers.
            random.seed (random.getrandbits (32) + m)
            
        if self.lastBound0 != (low0, high0):
            self.lastBound0 = (low0, high0)
            self.sources = []
            for i in self.mask.set0.boundedIterator (low0, high0):
                self.sources.append (i)

        nSources = len (self.sources)
        for j in self.mask.set1.boundedIterator (low1, high1):
            s = []
            for k in xrange (0, self.perTarget[m]):
                i = random.randint (0, self.N0 - 1)
                if i < nSources:
                    s.append (self.sources[i])
            s.sort ()
            for i in s:
                yield (i, j)
            m += 1

    def repr (self):
        return self._repr_applyop ('random(fanIn=%s)' % self.fanIn, self.mask)

    def _to_xml (self):
        return E ('apply',
                  E ('times'),
                  CSAObject.apply (FanInRandomOperator.tag, self.fanIn),
                  self.mask._to_xml ())


class FanOutRandomOperator (cs.Operator):
    tag = 'random_fanOut'
    
    def __init__ (self, fanOut):
        self.fanOut = fanOut

    def __mul__ (self, other):
        assert isinstance (other, cs.Finite) \
               and isinstance (other, cs.Mask), \
               'expected finite mask'
        return FanInRandomMask (self.fanOut, other.transpose ()).transpose ()

    def repr (self):
        return 'random(fanOut=%s)' % self.fanOut

    def _to_xml (self):
        return CSAObject.apply (FanOutRandomOperator.tag, self.fanOut)

registerTag (FanOutRandomOperator.tag, FanOutRandomOperator, 1)



#
#  This file is part of the Connection-Set Algebra (CSA).
#  Copyright (C) 2010,2011,2012 Mikael Djurfeldt
#
#  CSA is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  CSA is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math as _math
import random as _random
import numpy as _numpy

import intervalset as _iset

def grid2d (width, xScale = 1.0, yScale = 1.0, x0 = 0.0, y0 = 0.0):
    xScale /= width
    yScale /= width
    g = lambda i: \
        (x0 + xScale * (i % width), y0 + yScale * (i / width))
    g.type = 'grid'
    g.width = width
    g.xScale = xScale
    g.yScale = yScale
    g.x0 = x0
    g.y0 = y0
    g.inverse = lambda x, y: \
                    int (round (x / xScale - x0)) \
                    + width * int (round (y / yScale - y0))
    return g

def random2d (N, xScale = 1.0, yScale = 1.0):
    coords = [(xScale * _random.random (), yScale * _random.random ())
              for i in xrange (0, N)]
    g = lambda i: coords[i]
    g.type = 'ramdom'
    g.N = N
    g.xScale = xScale
    g.yScale = yScale
    # We should use a KD-tree here
    g.inverse = lambda x, y, domain=_iset.IntervalSet ((0, N - 1)): \
                    _numpy.array ([euclidDistance2d ((x, y), g(i)) \
                                   for i in domain]).argmin () \
                                   + domain.min ()
    return g

def euclidDistance2d (p1, p2):
    dx = p1[0] - p2[0]
    dy = p1[1] - p2[1]
    return _math.sqrt (dx * dx + dy * dy)

class ProjectionOperator (object):
    def __init__ (self, projection):
        self.projection = projection

    def __mul__ (self, g):
        projection = self.projection
        return lambda i: projection (g (i))

def euclidMetric2d (g1, g2 = None):
    g2 = g1 if g2 == None else g2
    return lambda i, j: euclidDistance2d (g1 (i), g2 (j))

#del random



#!/usr/bin/env python

from distutils.core import setup
from csa.version import __version__

long_description = """The CSA library provides elementary connection-sets and operators for
combining them. It also provides an iteration interface to such
connection-sets enabling efficient iteration over existing connections
with a small memory footprint also for very large networks. The CSA
can be used as a component of neuronal network simulators or other
tools."""

setup (
    name = "csa",
    version = __version__,
    packages = ['csa',],
    author = "Mikael Djurfeldt", # add your name here if you contribute to the code
    author_email = "mikael@djurfeldt.com",
    description = "The Connection-Set Algebra implemented in Python",
    long_description = long_description,
    license = "GPLv3",
    keywords = "computational neuroscience modeling connectivity",
    url = "http://software.incf.org/software/csa/",
    classifiers = ['Development Status :: 2 - Pre-Alpha',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Natural Language :: English',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 2',
                   'Topic :: Scientific/Engineering'],
    requires = ['numpy', 'matplotlib'],
)
